/*
 * Speculation control stuff
 *
 */
#include <linux/module.h>

#include <asm/msr.h>
#include <asm/processor.h>
#include <asm/spec_ctrl.h>

/*
 * Keep it open for more flags in case needed.
 *
 * -1 means "not touched by nospec() earlyparam"
 *
 *
 * If IBRS is set, IBPB is always set. IBPB can be set independently
 * on IBRS state (SKL).
 */
int ibrs_state = -1;
static int ibpb_state = -1;

unsigned int notrace x86_ibrs_enabled(void)
{
	if (ibrs_state != 1)
		return 0;
	return 1;
}
EXPORT_SYMBOL_GPL(x86_ibrs_enabled);

unsigned int notrace x86_ibpb_enabled(void)
{
	if (ibpb_state != 1)
		return 0;
	return 1;
}
EXPORT_SYMBOL_GPL(x86_ibpb_enabled);

void x86_disable_ibrs(void)
{
	if (x86_ibrs_enabled())
		native_wrmsrl(MSR_IA32_SPEC_CTRL, 0);
}
EXPORT_SYMBOL_GPL(x86_disable_ibrs);

void x86_enable_ibrs(void)
{
	if (x86_ibrs_enabled())
		native_wrmsrl(MSR_IA32_SPEC_CTRL, FEATURE_ENABLE_IBRS);
}
EXPORT_SYMBOL_GPL(x86_enable_ibrs);

void stuff_RSB(void)
{
	/*
	 * Do nothing: this is to avoid kABI breakage.
	 */
}
EXPORT_SYMBOL_GPL(stuff_RSB);

/*
 * Called after upgrading microcode, check CPUID directly.
 */
void x86_spec_check(void)
{

	if (ibpb_state == 0) {
		printk_once(KERN_INFO "IBRS/IBPB: disabled\n");
		return;
	}

	if (cpuid_edx(7) & BIT(26)) {
		if (ibrs_state == -1) {
			/* noone force-disabled IBRS */
			ibrs_state = 1;
			printk_once(KERN_INFO "IBRS: initialized\n");
		}
		printk_once(KERN_INFO "IBPB: initialized\n");
		ibpb_state = 1;

		setup_force_cpu_cap(X86_FEATURE_SPEC_CTRL);
	}

	if (boot_cpu_data.x86_vendor == X86_VENDOR_AMD) {
		if (cpuid_ebx(0x80000008) & BIT(12)) {
			ibpb_state = 1;
			printk_once(KERN_INFO "IBPB: Initialized\n");
		} else {
			u64 val;

			switch (boot_cpu_data.x86) {
			case 0x10:
			case 0x12:
			case 0x16:
				printk_once(KERN_INFO
					"IBPB: Disabling indirect branch predictor support\n");
				rdmsrl(MSR_F15H_IC_CFG, val);
				val |= MSR_F15H_IC_CFG_DIS_IND;
				wrmsrl(MSR_F15H_IC_CFG, val);
				break;
			}
			ibpb_state = 0;
		}
	}
}
EXPORT_SYMBOL_GPL(x86_spec_check);

int __init nospec(char *str)
{
	setup_clear_cpu_cap(X86_FEATURE_SPEC_CTRL);
	ibrs_state = 0;
	ibpb_state = 0;

	return 0;
}
early_param("nospec", nospec);
